#!/usr/bin/python
# coding=UTF-8

import subprocess
import re
import pdb

""" pdftotext -layout -nopgbrk p1220.pdf - """


class PDFRecordFinder(object):
    def __init__(self, src, heading_exp=None):
        if not heading_exp:
            heading_exp = re.compile('\s+Record Name: (.*)')

        opts = ["pdftotext", "-layout", "-nopgbrk", src, '-']
        pdftext = subprocess.check_output(opts)
        self.textrows = pdftext.split('\n')
        self.heading_exp = heading_exp

    """
    def columns(self):
        results = []
        cc = ColumnCollector()
        for heading, group in self.record_grouping():
            print "HEADING", heading
            for row in group:
                if cc.is_next_field(row):
                    yield cc
                    cc = ColumnCollector()
                #print row
                try:
                    cc.add(row)
                except UnknownColumn, e:
                    results.append(cc)
                    cc = ColumnCollector()



    def record_grouping(self):
        row_iter = iter(self.textrows)
        i = 0
        for row in row_iter:
            i += 1
            match = self.heading_exp.match(row)
            if match:
                yield (match.groups(), self.extract_record_columns(row_iter))

    """


    def locate_heading_rows(self):
        results = []
        for (i, row) in enumerate(self.textrows):
            match = self.heading_exp.match(row)
            if match:
                if not ''.join(match.groups()).lower().endswith('(continued)'):
                    results.append((i, ''.join(match.groups())))
        return results


    def records2(self):
        row_iter = iter(self.textrows)
        record = True
        while record:
            record = self.extract_record(row_iter)
            yield record


    def extract_record(self, row_iter):
        heading = self.find_heading(row_iter)
        fields = self.find_fields(row_iter)
        return heading, list(fields)
       

    def find_heading(self, row_iter):
        for row in row_iter:
            heading_match = self.heading_exp.match(row)
            if heading_match:
                return heading_match.groups()


    def find_fields(self, row_iter):
        cc = ColumnCollector()
        for r in row_iter:
            row = self.extract_columns_from_row(r)
            if not row:
                continue
            
            if cc.is_next_field(row):
                yield cc
                cc = ColumnCollector()
            try:
                cc.add(row)
            except UnknownColumn, e:
                print 'UNKNOWN COLUMN', row
                raise StopIteration


    def extract_columns_from_row(self, row):
        re_multiwhite = re.compile(r'\s{2,}')

        # IF LINE DOESN'T CONTAIN MULTIPLE WHITESPACES, IT'S LIKELY NOT A TABLE
        if not re_multiwhite.search(row):
            return None

        white_ranges = [0,]
        pos = 0
        while pos < len(row):
            match = re_multiwhite.search(row[pos:])
            if match:
                white_ranges.append(pos + match.start())
                white_ranges.append(pos + match.end())
                pos += match.end()
            else:
                white_ranges.append(len(row))
                pos = len(row)

        row_result = []
        white_iter = iter(white_ranges)
        while white_iter:
            try:
                start = white_iter.next()
                end = white_iter.next()
                if start != end:
                    row_result.append(
                        (start, row[start:end])
                    )
                
            except StopIteration:
                white_iter = None

        return row_result
         
    

    def extract_row_columns(self, row_iter):
        re_multiwhite = re.compile(r'\s{2,}')
        full_width_text_count = 0

        #for r in row_iter:
        row = None
        while not row:
            row = row_iter.next()
            row = row.decode('UTF-8')
        
            # IF LINE CONTAINS MULTIPLE WHITESPACE, IT IS PROBABLY A TABLE
            if not re_multiwhite.search(row):
                full_width_text_count += 1
                row = None

        if True:
            white_ranges = [0,]
            pos = 0
            match = True
            while pos < len(row):
                match = re_multiwhite.search(row[pos:])
                if match:
                    white_ranges.append(pos + match.start())
                    white_ranges.append(pos + match.end())
                    pos += match.end()
                else:
                    white_ranges.append(len(row))
                    pos = len(row)


            row_result = []
            white_iter = iter(white_ranges)
            while white_iter:
                try:
                    start = white_iter.next()
                    end = white_iter.next()
                    if start != end:
                        row_result.append(
                            (start, row[start:end])
                        )
                
                except StopIteration:
                    white_iter = None
            
            #print white_ranges
            return row_result
            #yield row_result
            #result.append(row_result)
             

            """
            row_result = []
            pos = 0
            while pos < len(row):
                match = re_multiwhite.search(row[pos:])
                if match and match.start() > 0:
                    row_result.append((
                        pos,
                        row[pos:pos+match.start()],))
                    pos += match.end()
                else:
                    if match:
                        pos += match.end()
                    row_result.append((pos,row[pos:]))
                    pos += len(row)
            result.append(row_result)
            row_result = []
            """
        #return result


class UnknownColumn(Exception):
    pass

class IsNextField(Exception):
    pass

class ColumnCollector(object):
    def __init__(self, initial=None):
        self.data = None
        pass

    def add(self, data):
        if not self.data:
            self.data = dict(data)
        else:
            if self.is_next_field(data):
                raise IsNextField()
            for col_id, value in data:
                self.merge_column(col_id, value)
                
    def merge_column(self, col_id, value):
        if col_id in self.data.keys():
            self.data[col_id] += ' ' + value.strip()
            
        else:        
            # try adding a wiggle room value?
            raise UnknownColumn
    
    def is_next_field(self, data):
        """
        If the first key value contains a string
        and we already have some data in the record,
        then this row is probably the beginning of
        the next field. Raise an exception and continue
        on with a fresh ColumnCollector.
        """
        first_key = dict(data).keys()[0]
        if self.data:
            return self.data.keys()[0] == first_key
        return False

        


